#!/usr/bin/which python
#
# Launch a juju api server with some sample data.
#
#
# author: kapilt
"""

TODO:

  Subordinate service unit handling.

    - data generator on new relation or new unit
    - check scope of relation if subordinate, add sub units
    - on new unit, check service relations for sub and add sub unit
    - on removed unit, check service relations for sub and remove sub unit.
"""

import argparse
import json
import logging
import os
import random
import shutil
import tempfile
import yaml
import zookeeper

from twisted.internet.defer import Deferred, inlineCallbacks, succeed
from twisted.internet import reactor

from juju.agents.api import APIEndpointAgent
from juju.charm.publisher import CharmPublisher
from juju.charm.repository import (
    CharmURL, RemoteCharmRepository, CS_STORE_URL)

from juju.environment.config import EnvironmentsConfig
from juju.errors import JujuError
from juju.state.environment import EnvironmentStateManager

from juju.machine.constraints import ConstraintSet
from juju.providers.dummy import MachineProvider, DummyMachine
from juju.providers.common.files import FileStorage

from juju.state.base import StateBase
from juju.state.endpoint import RelationEndpoint
from juju.state.errors import StopWatcher
from juju.state.initialize import StateHierarchy
from juju.state.machine import MachineStateManager
from juju.state.placement import _unassigned_placement
from juju.state.relation import RelationStateManager
from juju.state.security import Principal
from juju.state.service import ServiceStateManager
from juju.state.utils import get_open_port, YAMLState
from juju.tests.common import zookeeper_test_context



from txzookeeper.managed import ManagedClient

log = logging.getLogger("jitsu.scale")


JUJU_ENV_CONF = """
environments:
  dervish:
    type: dummy
    default-series: precise
"""


class TempDir(object):

    def __enter__(self):
        self.temp_dir = tempfile.mkdtemp()
        return self.temp_dir

    def __exit__(self, exc_type, exc_value, traceback):
        if os.path.exists(self.temp_dir):
            shutil.rmtree(self.temp_dir)


class Machine(DummyMachine):

    def __init__(self, instance_id, dns_name, private_dns, state):
        self.instance_id = instance_id
        self.dns_name = dns_name
        self.private_dns_name = private_dns
        self.state = state


class MachineProvider(MachineProvider):

    def start_machine(self, machine_data, master=False):
        """Start a machine in the provider."""

        machine = Machine(
            machine_data['machine_id'],
            machine_data['dns_name'],
            machine_data['priv_dns_name'],
            machine_data['state']
            )
        self._machines.append(machine)
        return succeed([machine])


class DataGenerator(StateBase):

    # weighted
    unit_states = (
        'pending', 'installed', 'install_error',
        'started', 'started', 'started', 'started',
        'start_error', 'stopped', 'stop_error')

    # weighted / 80% success
    relation_states = ('up', 'up', 'up', 'down', 'down')

    # weighted / 90% running
    machine_agent_states = (
        'pending', 'running', 'running', 'running')

    machine_states = (
        'running', 'pending', 'terminated', 'stopped')

    def __init__(self, client, provider):
        super(DataGenerator, self).__init__(client)
        self.address_block = "192.168.%s.%s"
        self.provider = provider
        self.addresses = set()
        self.seen = set()
        self.running = False
        self.finished = Deferred()

    @inlineCallbacks
    def init_service_unit(self, topo, service_id, unit_id, unit_name):

        relations = topo.get_relations_for_service(service_id)

        # Topology doesn't store subordinate status, this is a best
        # approximation, technically a subordinate can be the server.
        is_sub = bool(
            [r for r in relations if r['scope'] == 'container' \
             and r['service']['role'] == 'client'])

        # Find the machine for the unit.
        if is_sub:
            container_id = topo.get_service_unit_principal(unit_id)
            assert container_id, "no container for subordinate %s" % unit_name
            container_svc_id = topo.get_service_unit_service(container_id)
            machine_id = topo.get_service_unit_machine(
                container_svc_id, container_id)
        else:
            machine_id = topo.get_service_unit_machine(service_id, unit_id)

        log.debug("Processing new unit %s:%s (sub: %s) on machine %s" % (
            unit_name, unit_id, is_sub, machine_id))

        # If the unit's machine is pending, bail.
        if machine_id is None:
            return

        machine = yield self.provider.get_machine(machine_id)
        if machine.state == 'pending':
            return

        self.seen.add(unit_id)

        # TODO for subs check principal is running.
        state = random.choice(self.unit_states)
        if state == 'pending':
            return

        node = YAMLState(self._client, "/units/%s" % unit_id)
        yield node.read()

        unit_address = self._new_address()

        w = {unit_name: yaml.safe_dump(
            {'state': state, 'state_variables': {}})}

        for r in relations:
            # simulate the unit's relation state enough for status.
            # TODO: subordinate relation state
            if r['scope'] == 'container':
                continue
            alive_path = "/relations/%s/%s/%s" % (
                r['relation_id'], r['service']['role'], unit_id)
            yield self._client.create(alive_path)

            w[r['relation_id']] = yaml.safe_dump({
                'state': random.choice(self.relation_states),
                'state_variables': {}})

        d = {'public-address': unit_address,
             'private-address': unit_address,
             'workflow_state': w}

        node.update(d)

        try:
            yield self._client.create("/units/%s/agent" % unit_id, "")
        except zookeeper.NodeExistsException:
            log.debug("Unit: %s already has agent.." % unit_id)
            pass

        yield node.write()
        n, s = yield self._client.get("/units/%s" % unit_id)

    @inlineCallbacks
    def init_machine(self, mid):
        maddr = self._new_address()
        md = {'machine_id': mid, 'priv_dns_name': maddr,
              'state': 'running', 'dns_name': maddr}
        yield self.provider.start_machine(md)
        node = YAMLState(self._client, "/machines/%s" % mid)
        yield node.read()
        node['provider-machine-id'] = mid
        yield node.write()
        if md['state'] == 'running':
            yield self._client.create("/machines/%s/agent" % mid, "")

    def _new_address(self):
        while 1:
            addr = self.address_block % (
                random.randint(1, 256), random.randint(1, 256))
            if addr in self.addresses:
                continue
            self.addresses.add(addr)
            return addr

    @inlineCallbacks
    def process(self, old, new):
        """
        """
        log.info("DataGen processing topology delta")
        if not self.running:
            self.finished.callback(True)
            raise StopWatcher()

        # Process new and removed machines.
        cur_machines = set(new.get_machines())
        if old is None:
            old_machines = set()
        else:
            old_machines = set(old.get_machines())

        for m in (cur_machines - old_machines):
            #log.info("Adding machine %s" % m)
            yield self.init_machine(m)

        removed = list(old_machines - cur_machines)
        p_machines = yield self.provider.get_machines(removed)
        yield self.provider.shutdown_machines(p_machines)

        # Process new units.
        cur_services = set(new.get_services())
        cur_units = set()
        for s in cur_services:
            get_unit_name = lambda uid: (
                s, uid, new.get_service_unit_name(s, uid))
            cur_units.update(
                map(get_unit_name, new.get_service_units(s)))

        for su in cur_units:
            if su[1] in self.seen:
                continue
            #log.info("Adding unit %s" % (su,))
            yield self.init_service_unit(new, *su)

        self.count += 1

    def start(self):
        assert not self.running
        self.running = True
        self.count = 0
        self.finished = Deferred()
        return self._watch_topology(self.process)

    def stop(self):
        if not self.running:
            return
        self.running = False
        return self.finished


@inlineCallbacks
def load(client, data):
    """load up an environment export (based on jitsu import)
    """
    # Get state managers
    services = ServiceStateManager(client)
    relations = RelationStateManager(client)
    machines = MachineStateManager(client)

    # Resolve charms from charm store
    store = RemoteCharmRepository(CS_STORE_URL)
    charms = []
    for s in data['services']:
        curl = CharmURL.infer(s['charm'], "precise")
        if s['charm'].startswith('local:'):
            raise JujuError("Local charm not supported")
        else:
            assert s['charm'].startswith('cs:')
            charm = yield store.find(curl)
            charms.append((s['charm'], charm))

    # Upload charms to the environment
    with TempDir() as temp_dir:
        storage = FileStorage(temp_dir)
        publisher = CharmPublisher(client, storage)
        seen = set()
        for cid, c in charms:
            if cid in seen:
                continue
            seen.add(cid)
            log.info("Publishing charm %s", cid)
            yield publisher.add_charm(
                str(CharmURL.infer(cid, "precise")),
                c)
        charm_states = yield publisher.publish()

    # Index by url
    charm_map = dict([(ci[0], cs) for ci, cs in zip(charms, charm_states)])
    constraint_set = ConstraintSet("dummy")
    constraint_set.register_generics([])

    # Create services
    for s in data['services']:
        log.info("Creating service %s %s", s['name'], charm_map[s['charm']].id)
        constraints = constraint_set.load(s['constraints'])
        svc = yield services.add_service_state(
            s['name'],
            charm_map[s['charm']],
            constraints)
        config = yield svc.get_config()
        config.update(s['config'])
        yield config.write()

        # Create units
        log.info("Creating units %d", s['unit_count'])
        for i in range(s['unit_count']):
            unit = yield svc.add_unit_state()
            yield _unassigned_placement(client, machines, unit)

    # Add relations
    for r in data['relations']:
        if len(r) == 1:
            eps = [RelationEndpoint(*r[0])]
        else:
            eps = [RelationEndpoint(*r[0]), RelationEndpoint(*r[1])]
        log.info("Adding relation %s" % (
            " ".join(map(
                lambda r: "%s:%s %s" % (
                    r.service_name, r.relation_name, r.relation_role),
                eps))))
        yield relations.add_relation_state(*eps)


@inlineCallbacks
def initialize(api_port, zk_port, export_data, agent_dir):

    # Setup the state hierarchy.
    client = yield ManagedClient("localhost:%s" % zk_port).connect()
    principal = Principal("admin", "admin")
    constraints = {'arch': 'amd64', 'cpu': 1,
                   'ubuntu-series': 'precise', 'provider-type': 'dummy'}
    log.info("Initializing state")
    state = StateHierarchy(
        client, principal.get_token(), "0", constraints, "dummy")
    yield state.initialize()

    # Initialize the environment config state.
    config = EnvironmentsConfig()
    config.parse(JUJU_ENV_CONF)
    env = EnvironmentStateManager(client)
    yield env.set_config_state(config, "dervish")

    provider = MachineProvider("dervish", config.get_default())

    # Attach the client
    yield principal.attach(client)

    # Load the tree
    yield load(client, export_data)

    # Hook up the runtime data simulator
    #   Generates appropriate state for units, and machines.
    generator = DataGenerator(client, provider)
    yield generator.start()

    # Start the api agent
    agent = APIEndpointAgent()
    agent.configure({
        "zookeeper_servers": "localhost:%s" % zk_port,
        "logfile": "-",
        "juju_directory": agent_dir,
        "port": api_port,
        "loglevel": "DEBUG",
        "session_file": os.path.join(agent_dir, "session.yml")
        })
    log.info("Starting api agent..")
    yield agent.startService()


@inlineCallbacks
def run_one(func, *args, **kw):
    try:
        yield func(*args, **kw)
    except:
        import traceback
        traceback.print_exc()
        reactor.stop()


def get_parser():
    parser = argparse.ArgumentParser(
        description="Start an api server with data")
    parser.add_argument("-p", "--port", default=8081, type=int,
                        help="Port to use")
    parser.add_argument("-f", "--file", required=True,
                        type=argparse.FileType('r'),
                        help="Environment json export")

    return parser


def main():
    parser = get_parser()
    options = parser.parse_args()
    log_options = {
        "level": logging.DEBUG,
        "format": "%(asctime)s %(name)s:%(levelname)s %(message)s"}

    logging.basicConfig(**log_options)
    zookeeper.set_debug_level(0)
    export_data = json.loads(options.file.read())
    zk_port = get_open_port()


    if not "ZOOKEEPER_PATH" in os.environ:
        if not os.path.exists('/etc/zookeeper/conf'):
            raise ValueError("ZK not found")
        os.environ['ZOOKEEPER_PATH'] = "system"

    with TempDir() as agent_dir:
        with zookeeper_test_context(os.environ['ZOOKEEPER_PATH'], zk_port):
            reactor.callWhenRunning(
                run_one, initialize,
                options.port, zk_port, export_data, agent_dir)
            reactor.run()

if __name__ == '__main__':
    main()

