from twisted.internet.defer import inlineCallbacks, Deferred
from twisted.internet.threads import deferToThread

from juju.errors import ServiceError

import os
import subprocess


def _check_call(args, env=None, output_path=None):
    if not output_path:
        f = open(os.devnull, "a")
    else:
        check_path = os.path.exists(output_path) and output_path \
                     or os.path.dirname(output_path)
        if os.access(check_path, os.W_OK):
            f = open(output_path, "a")
        else:
            raise ValueError(
                "Output path inaccessible %s" % output_path)
    try:
        return subprocess.check_call(
            args,
            stdout=f.fileno(), stderr=f.fileno(),
            env=env)
    finally:
        f.close()


def _cat(filename, use_sudo=False):
    args = ("cat", filename)
    if use_sudo and not os.access(filename, os.R_OK):
        args = ("sudo",) + args
    elif os.path.exists(filename):
        with open(filename) as fh:
            return (0, fh.read())
    p = subprocess.Popen(
        args, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
    stdout_data, _ = p.communicate()
    r = p.returncode
    return (r, stdout_data)


class TwistedDaemonService(object):
    """Manage the starting and stopping of an Agent.

    This manager tracks the agent via its --pidfile. The pidfile argument
    specifies the location of the pid file that is used to track this service.

    """
    def __init__(self, name, pidfile, output_path, use_sudo=False):
        self._name = name
        self._use_sudo = use_sudo
        self._description = None
        self._environ = None
        self._command = None
        self._daemon = True
        self.output_path = output_path
        self._pid_path = pidfile
        self._pid = None

    def set_description(self, description):
        self._description = description

    def set_daemon(self, value):
        self._daemon = bool(value)

    def set_environ(self, environ):
        for k, v in environ.items():
            environ[k] = str(v)
        self._environ = environ

    def set_output_path(self, output_path):
        self._output_path = output_path

    def set_command(self, command):
        if self._daemon:
            if "--pidfile" not in command:
                command += ["--pidfile", self._pid_path]
            else:
                # pid file is in command (consume it for get_pid)
                idx = command.index("--pidfile")
                self._pid_path = command[idx + 1]
        self._command = command

    @inlineCallbacks
    def _trash_output(self):
        if self.output_path and os.path.exists(self.output_path):
            # Just using os.unlink will fail when we're running TEST_SUDO
            # tests which hit this code path (because root will own
            # self.output_path)
            yield self._call("rm", "-f", self.output_path)

        if os.path.exists(self._pid_path):
            yield self._call("rm", "-f", self._pid_path)

    def _call(self, *args, **kwargs):
        # sudo even with -E will strip pythonpath so pass it to the command.
        if self._use_sudo:
            if self._environ:
                _args = ["%s=%s" % (k, v) for k, v in self._environ.items()]
            else:
                _args = []
            _args.insert(0, "sudo")
            _args.extend(args)
            args = _args

        return deferToThread(_check_call, args, env=self._environ,
                             output_path=self.output_path)

    @inlineCallbacks
    def start(self):
        if (yield self.is_running()):
            return
            raise ServiceError(
                "%s already running: pid (%s)" % (
                    self._name, self.get_pid()))
        yield self._trash_output()
        yield self._call(*self._command, output_path=self.output_path)
        yield self._sleep(0.2)

    def _sleep(self, delay):
        """Non-blocking sleep."""
        from twisted.internet import reactor
        deferred = Deferred()
        reactor.callLater(delay, deferred.callback, None)
        return deferred

    @inlineCallbacks
    def destroy(self):
        if not (yield self.is_running()):
            return
        yield self._call("kill", str(self.get_pid()))
        yield self._trash_output()

    def get_pid(self):
        if self._pid != None:
            return self._pid

        if not os.path.exists(self._pid_path):
            return None
        r, data = _cat(self._pid_path, use_sudo=self._use_sudo)
        if r != 0:
            return None

        data = data.strip()
        if not data.isdigit():
            return None
        self._pid = int(data)
        return self._pid

    def is_running(self):
        pid = self.get_pid()
        if not pid:
            return False
        proc_file = "/proc/%s" % pid
        if not os.path.exists(proc_file):
            return False
        return True
