from twisted.internet.defer import inlineCallbacks, returnValue

from juju.charm.errors import InvalidCharmHook
from juju.state.charm import CharmStateManager
from juju.state.service import ServiceStateManager


@inlineCallbacks
def debug_hooks(context, unit_name, hook_names):

    manager = ServiceStateManager(context.client)
    unit = yield manager.get_unit_state(unit_name)

    # Verify hook name
    if hook_names != ["*"]:
        context.log.debug("Verifying hook names...")
        yield validate_hooks(context.client, unit, hook_names)

    # Enable debug hooks
    context.log.debug(
        "Enabling hook debug on unit (%r)..." % unit_name)
    yield unit.enable_hook_debug(hook_names)

    # Ensure the unit is up.
    context.log.info("Waiting for unit")

    # Wait and verify the agent is running.
    while 1:
        exists_d, watch_d = unit.watch_agent()
        exists = yield exists_d
        if exists:
            context.log.info("Unit running")
            break
        yield watch_d


@inlineCallbacks
def validate_hooks(client, unit_state, hook_names):

    # Assemble a list of valid hooks for the charm.
    valid_hooks = ["start", "stop", "install", "config-changed"]
    service_manager = ServiceStateManager(client)
    endpoints = yield service_manager.get_relation_endpoints(
        unit_state.service_name)
    endpoint_names = [ep.relation_name for ep in endpoints]
    for endpoint_name in endpoint_names:
        valid_hooks.extend([
            endpoint_name + "-relation-joined",
            endpoint_name + "-relation-changed",
            endpoint_name + "-relation-departed",
            endpoint_name + "-relation-broken",
        ])

    # Verify the debug names.
    for hook_name in hook_names:
        if hook_name in valid_hooks:
            continue
        break
    else:
        returnValue(True)

    # We dereference to the charm to give a fully qualified error
    # message.  I wish this was a little easier to dereference, the
    # service_manager.get_relation_endpoints effectively does this
    # already.
    service_manager = ServiceStateManager(client)
    service_state = yield service_manager.get_service_state(
        unit_state.service_name)
    charm_id = yield service_state.get_charm_id()
    charm_manager = CharmStateManager(client)
    charm = yield charm_manager.get_charm_state(charm_id)
    raise InvalidCharmHook(charm.id, hook_name)
