from twisted.internet.defer import inlineCallbacks, returnValue

from juju.charm.errors import CharmNotFound, RepositoryNotFound
from juju.charm.publisher import CharmPublisher
from juju.charm.repository import resolve
from juju.lib import serializer

from juju.state.endpoint import RelationEndpoint
from juju.state.placement import place_unit
from juju.state.relation import RelationStateManager
from juju.state.service import ServiceStateManager


@inlineCallbacks
def deploy(context,
           charm_name,
           service_name=None,
           constraint_strs=(),
           config=None, config_raw=None, num_units=1):
    """Deploy a charm within an environment.

    This will publish the charm to the environment, creating
    a service from the charm, and get it set to be launched
    on a new machine. If --repository is not specified, it
    will be taken from the environment variable JUJU_REPOSITORY.
    """

    # Locate charm, no local repos yet.
    try:
        store, charm_url = resolve(
            charm_name, "/dev/null", context.default_series)
    except RepositoryNotFound:
        raise CharmNotFound("charm store", charm_name)

    store.cache_path = context.charm_cache_dir

    charm = yield store.find(charm_url)

    # Validate config options prior to deployment attempt
    service_name = service_name or charm_url.name
    service_options = {}
    if config_raw:
        config = serializer.yaml_load(config_raw)
    if config:
        service_options = charm.config.validate(config)

    charm_id = str(charm_url.with_revision(charm.get_revision()))
    context.log.info("Deploying service %s using charm %s" % (
        service_name, charm_id))
    constraint_set = yield context.provider.get_constraint_set()
    placement_policy = context.provider.get_placement_policy()
    constraints = constraint_set.parse(constraint_strs)

    # Publish the charm to juju
    storage = yield context.provider.get_file_storage()
    publisher = CharmPublisher(context.client, storage)
    yield publisher.add_charm(charm_id, charm)
    result = yield publisher.publish()

    # In future we might have multiple charms be published at
    # the same time.  For now, extract the charm_state from the
    # list.
    charm_state = result[0]

    # Create the service state
    service_manager = ServiceStateManager(context.client)
    service_state = yield service_manager.add_service_state(
        service_name, charm_state, constraints)

    # Use the charm's ConfigOptions instance to validate service
    # options.. Invalid options passed will thrown an exception
    # and prevent the deploy.
    state = yield service_state.get_config()
    charm_config = yield charm_state.get_config()

    # Return the validated options with the defaults included.
    service_options = charm_config.validate(service_options)
    state.update(service_options)
    yield state.write()

    # Create desired number of service units
    if (yield service_state.is_subordinate()):
        context.log.info("Subordinate %r awaiting relationship "
                 "to principal for deployment.", service_name)
    else:
        for i in xrange(num_units):
            unit_state = yield service_state.add_unit_state()
            yield place_unit(
                context.client, placement_policy, unit_state)

    # Check if we have any peer relations to establish
    if charm.metadata.peers:
        relation_manager = RelationStateManager(context.client)
        for peer_name, peer_info in charm.metadata.peers.items():
            yield relation_manager.add_relation_state(
                RelationEndpoint(service_name,
                                 peer_info["interface"],
                                 peer_name,
                                 "peer"))

    context.log.info("Charm deployed as service: %r", service_name)
    returnValue(True)
