from twisted.internet.defer import inlineCallbacks

from juju.charm.publisher import CharmPublisher
from juju.charm.repository import (
    CharmURL, RemoteCharmRepository, CS_STORE_URL)

from juju.errors import JujuError

from juju.state.endpoint import RelationEndpoint
from juju.state.service import ServiceStateManager
from juju.state.machine import MachineStateManager
from juju.state.placement import _unassigned_placement
from juju.state.relation import RelationStateManager


@inlineCallbacks
def load(context, data):
    """

    Check for name conflicts
     - Upload charms
     - Create services (+ config, constraints)
     - Add units
     - Add relations
    """
    # Get state managers
    services = ServiceStateManager(context.client)
    relations = RelationStateManager(context.client)
    machines = MachineStateManager(context.client)

    # First detect conflicts
    existing_services = yield services.get_all_service_states()
    env_svc_names = [e.service_name for e in existing_services]
    import_svc_names = [i['name'] for i in data['services']]
    conflict_svc_names = set(env_svc_names) & set(import_svc_names)
    if conflict_svc_names:
        raise JujuError(
            "Import has name conflicts with existing services %s" % (
                ", ".join(conflict_svc_names)))

    store = RemoteCharmRepository(CS_STORE_URL)

    charms = []
    for s in data['services']:
        curl = CharmURL.infer(s['charm'], context.default_series)
        if s['charm'].startswith('local:'):
            raise JujuError(
                "Local charm needed but repository not specified")
        else:
            assert s['charm'].startswith('cs:')
            charm = yield store.find(curl)
            charms.append((s['charm'], charm))

    # Upload charms to the environment
    publisher = CharmPublisher(
        context.client, context.provider.get_file_storage())
    for cid, c in charms:
        context.log.info("Publishing charm %s", cid)
        yield publisher.add_charm(
            str(CharmURL.infer(cid, context.default_series)),
            c)
    charm_states = yield publisher.publish()

    # Index by url
    charm_map = dict([(ci[0], cs) for ci, cs in zip(charms, charm_states)])
    constraint_set = yield context.provider.get_constraint_set()

    # Create services
    for s in data['services']:
        context.log.info(
            "Creating service %s %s", s['name'], charm_map[s['charm']].id)
        constraints = constraint_set.load(s['constraints'])
        svc = yield services.add_service_state(
            s['name'],
            charm_map[s['charm']],
            constraints)
        config = yield svc.get_config()
        config.update(s['config'])
        yield config.write()

        # Create units
        context.log.info("Creating units")
        for i in range(s['unit_count']):
            unit = yield svc.add_unit_state()
            yield _unassigned_placement(context.client, machines, unit)

    # Add relations
    for r in data['relations']:
        if len(r) == 1:
            eps = [RelationEndpoint(*r[0])]
        else:
            eps = [RelationEndpoint(*r[0]), RelationEndpoint(*r[1])]
        context.log.info("Adding relation %s" % (
            " ".join(map(
                lambda r: "%s:%s %s" % (
                    r.service_name, r.relation_name, r.relation_role),
                eps))))
        yield relations.add_relation_state(*eps)
