from twisted.internet.defer import inlineCallbacks, returnValue

from juju.state.errors import (AmbiguousRelation, NoMatchingEndpoints,
                               UnsupportedSubordinateServiceRemoval)
from juju.state.relation import RelationStateManager
from juju.state.service import ServiceStateManager


@inlineCallbacks
def remove_relation(context, descriptors=()):

    relation_state_manager = RelationStateManager(context.client)
    service_state_manager = ServiceStateManager(context.client)
    endpoint_pairs = yield service_state_manager.join_descriptors(
            *descriptors)

    if len(endpoint_pairs) == 0:
        raise NoMatchingEndpoints()
    elif len(endpoint_pairs) > 1:
        raise AmbiguousRelation(descriptors, endpoint_pairs)

    # At this point we just have one endpoint pair. We need to pick
    # just one of the endpoints if it's a peer endpoint, since that's
    # our current API - join descriptors takes two descriptors, but
    # add_relation_state takes one or two endpoints. TODO consider
    # refactoring.
    endpoints = endpoint_pairs[0]
    if endpoints[0] == endpoints[1]:
        endpoints = endpoints[0:1]
    relation_state = yield relation_state_manager.get_relation_state(
        *endpoints)

    # Look at both endpoints, if we are dealing with a container relation
    # decide if one end is a principal.
    service_pair = []  # ordered such that sub, principal

    is_container = False
    has_principal = True
    for ep in endpoints:
        if ep.relation_scope == "container":
            is_container = True
        service = yield service_state_manager.get_service_state(
            ep.service_name)
        if (yield service.is_subordinate()):
            service_pair.append(service)
            has_principal = True
        else:
            service_pair.insert(0, service)
    if is_container and len(service_pair) == 2 and has_principal:
        sub, principal = service_pair
        raise UnsupportedSubordinateServiceRemoval(sub.service_name,
                                                   principal.service_name)

    yield relation_state_manager.remove_relation_state(relation_state)
    context.log.info("Removed %s relation from all service units.",
                     endpoints[0].relation_type)

    returnValue(True)
