from StringIO import StringIO
from twisted.internet.defer import maybeDeferred, succeed

from juju.rapi.cmd import add_relation
from juju.rapi.cmd import add_unit
from juju.rapi.cmd import config_get
from juju.rapi.cmd import config_set
from juju.rapi.cmd import constraints_set
from juju.rapi.cmd import constraints_get
from juju.rapi.cmd import deploy
from juju.rapi.cmd import debug_hooks
from juju.rapi.cmd import destroy_service
from juju.rapi.cmd import expose
from juju.rapi.cmd import export
from juju.rapi.cmd import import_env
from juju.rapi.cmd import remove_relation
from juju.rapi.cmd import remove_unit
from juju.rapi.cmd import resolved
from juju.rapi.cmd import status
from juju.rapi.cmd import terminate_machine
from juju.rapi.cmd import unexpose

from juju.rapi import rest

import logging

log = logging.getLogger('juju.rapi.context')


class LogBuffer(object):

    def __init__(self):
        self.msgs = []

    def debug(self, msg, *args):
        self.msgs.append(('debug', msg % args))

    def error(self, msg, *args):
        self.msgs.append(('error', msg % args))

    def exception(self, msg, *args):
        self.msgs.append(('error', msg % args))

    def info(self, msg, *args):
        self.msgs.append(('info', msg % args))

    def warning(self, msg, *args):
        self.msgs.append(('warning', msg % args))

    def reset(self):
        msgs = list(self.msgs)
        self.msgs = []
        return msgs


class APIContext(object):

    default_series = "precise"
    charm_cache_dir = "/tmp"

    def __init__(self, client, provider):
        self.client = client
        self.provider = provider
        self.log = LogBuffer()

    def connect(self, *args, **kw):
        if self.client.connected:
            return succeed(self)
        d = self.client.connect(*args, **kw)
        d.addCallback(lambda x: self)
        return d

    def close(self):
        if self.client.connected:
            self.client.close()

    def _invoke(self, func, *args, **kw):
        d = maybeDeferred(func, *args, **kw)
        return d.addCallback(
            lambda r: {'result': r, 'log': self.log.reset()}
            ).addErrback(self._on_error)

    def _on_error(self, failure):
        io = StringIO()
        failure.printTraceback(file=io)
        return {'err': True, 'log': self.log.reset(),
                'failure': io.getvalue()}

    def add_relation(self, endpoint_a, endpoint_b):
        """Add a relation to a service(s).
        """
        return self._invoke(
            add_relation.add_relation,
            self,
            endpoint_a,
            endpoint_b)

    def add_unit(self, service_name, num_units):
        """Add a unit to a service.
        """
        return self._invoke(
            add_unit.add_unit,
            self,
            service_name,
            num_units)

    def get_service(self, service_name):
        """Get information about the named service."""
        return self._invoke(
            rest.get_service,
            self,
            service_name)

    def get_endpoints(self, service_names=()):
        """Return available relation endpoints for the given services."""
        return self._invoke(
            rest.get_endpoints,
            self,
            service_names)

    def get_charm(self, charm_url):
        """Get information about the named charm from the env."""
        return self._invoke(
            rest.get_charm,
            self,
            charm_url)

    def get_constraints(self, named_entities=()):
        """Get the constraints of the named entities.

        If no entities are specified, return the constraints
        on the environment.
        """
        return self._invoke(
            constraints_get.constraints_get,
            self,
            named_entities
            )

    def get_config(self, service_name):
        """Get configuration of a service.
        """
        return self._invoke(
            config_get.config_get,
            self,
            service_name)

    def deploy(self,
               charm_url,
               service_name=None,
               charm_bundle=None,
               constraints=(),
               config=None,
               config_raw=None,
               num_units=1):
        """Deploy a charm as a service
        """
        return self._invoke(
            deploy.deploy,
            self,
            charm_url,
            service_name,
            constraints,
            config,
            config_raw,
            num_units
            )

    def debug_hooks(self, unit_name, hook_names=()):
        """Enable hook debug on a unit.
        """
        return self._invoke(
            debug_hooks.debug_hooks,
            self,
            unit_name,
            hook_names)

    def destroy_service(self, service_name):
        """Destroy the named service.
        """
        return self._invoke(
            destroy_service.destroy_service,
            self,
            service_name)

    def export(self):
        """Export an environment.
        """
        return self._invoke(
            export.export,
            self)

    def expose(self, service_name):
        """Expose a service for external access.
        """
        return self._invoke(
            expose.expose,
            self,
            service_name)

    def import_env(self, data):
        """Import an environment from data serialization.
        """
        return self._invoke(
            import_env.import_env,
            self,
            data)

    def remove_relation(self, endpoint_a, endpoint_b):
        """Remove a relation from a service(s).
        """
        return self._invoke(
            remove_relation.remove_relation,
            self,
            (endpoint_a, endpoint_b))

    def remove_units(self, unit_names):
        """Remove the given set of units.
        """
        return self._invoke(
            remove_unit.remove_unit,
            self,
            unit_names)

    def resolved(self, unit_name, relation_name=None, retry=False):
        """Mark an error as fixed.
        """
        return self._invoke(
            resolved.resolved,
            self,
            unit_name,
            relation_name,
            retry)

    def set_config(self, service_name, config, data=None):
        """Set a service's configuration.
        """
        return self._invoke(
            config_set.config_set,
            self,
            service_name,
            config,
            data)

    def set_constraints(self, service_name, constraints):
        """Set constraints on the environment or service.
        """
        return self._invoke(
            constraints_set.constraints_set,
            self,
            service_name,
            constraints)

    def status(self, scope=None):
        """Get information about the environment and its contained items.
        """
        return self._invoke(
            status.status,
            self,
            scope)

    def terminate_machine(self, machine_ids):
        """Terminate a machine.
        """
        return self._invoke(
            terminate_machine.terminate_machine,
            self,
            machine_ids)

    def unexpose(self, service_name):
        """Remove external access from a service.
        """
        return self._invoke(
            unexpose.unexpose,
            self,
            service_name)

    def upgrade_charm(self, service_name, local_charm=None):
        """Upgrade a charm.
        """
        raise NotImplementedError()

    def resolve_address(self, entity):
        """Given an entity return its public address.
        """
        raise NotImplementedError()
