"""
Object Stream Delta
 - status parsing atm
 - topology usage for relation comparison
"""
import copy
import logging
from twisted.internet.defer import inlineCallbacks, returnValue

# For unambigious relations deltas/identities.
from juju.state.topology import InternalTopology

log = logging.getLogger("juju.rapi.delta")


def _unit_key(name):
    name, number = name.split('/', 1)
    return (name, int(number))


class DeltaStreamManager(object):

    def __init__(self, context):
        self.context = context
        self.previous = {
            'services': {},
            'machines': {},
            'topology': None}
        self.streams = []  # protocol instances
        self.running = False

    def start(self):
        if self.running:
            return

        d = self.context.connect()

        def _start(ctx):
            self.running = True
            return self.pump()

        return d.addCallback(_start)

    def stop(self):
        self.running = False
        self.context.close()

    @inlineCallbacks
    def pump(self, *ignored):
        """Process state changes into notifications.
        """
        if not self.running:
            return

        try:
            current = (yield self.context.status())['result']
        except Exception, e:
            # The delta pump must go on.
            log.exception("An error occurred %s" % str(e))
            self._schedule()
            return

        current['topology'] = yield self._get_topology()

        changes = []
        delta(copy.deepcopy(self.previous),
              copy.deepcopy(current),
              changes.append)
        self.previous = current
        self.flush(changes)
        self._schedule()

    def _schedule(self):
        """Schedule the next event stream pump.
        """
        if not self.running:
            return

        from twisted.internet import reactor
        reactor.callLater(5, self.pump)

    @inlineCallbacks
    def _get_topology(self):
        """Delta streams relation resolution needs this.
        """
        t = InternalTopology()
        content, stat = yield self.context.client.get('/topology')
        t.parse(content)
        returnValue(t)

    def add(self, p):
        """Add a stream/protocol
        """
        self.streams.append(p)

    def remove(self, p):
        """Add a stream/protocol
        """
        if p in self.streams:
            self.streams.remove(p)

    def get_current(self):
        """Return current state as list of changes."""
        changes = []
        delta(
            {'machines': {}, 'services': {}, 'topology': None},
            copy.deepcopy(self.previous), changes.append)
        return changes

    def get_status(self):
        """Return the current status output."""
        status = dict(self.previous)
        del status['topology']
        return status

    def flush(self, delta):
        if not delta:
            return None
        for s in self.streams:
            s.publish_changes(delta)


def delta(prev, cur, stream):
    """
    Given two topology augmented status dictionaries, generate a
    stream of object events corresponding to the delta between the
    two states.

    Logical order matters wrt to the event stream.

    Bring Up / Bring down roughly reverse.
    --------
    Services
    Relations
    Machines
    Units


    TODO: A more efficient alternative to 1-1 delta changes, is
    assuming intelligence on the client, and only sending logical
    changes, ala sending remove service automatically removes units.
    ie. operational transform.
    """

    ## Services
    cur_svc_ids = set(cur['services'])
    prev_svc_ids = set(prev['services'])

    ## Relations
    if not prev.get('topology'):
        prev_rel_ids = set()
    else:
        prev_rel_ids = set(prev['topology'].get_relations())

    if cur['topology']:
        cur_rel_ids = set(cur['topology'].get_relations())
    else:
        cur_rel_ids = set()

    ## Units
    prev_units = {}
    for s in prev['services']:
        units = prev['services'][s].pop('units', ())
        prev_units.update(units)
        del units

    cur_units = {}
    for s in cur['services']:
        units = cur['services'][s].pop('units', ())
        cur_units.update(units)
        del units

    ## Machines
    prev_machine_ids = set(prev['machines'].keys())
    cur_machine_ids = set(cur['machines'].keys())

    # svc add
    for sid in cur_svc_ids - prev_svc_ids:
        svc = dict(cur['services'][sid])
        svc.pop('units', None)
        svc.pop('relations', None)
        svc['id'] = sid
        stream(('service', 'add', svc))

    # machine change
    for mid in cur_machine_ids.intersection(prev_machine_ids):
        if cur['machines'][mid] == prev['machines'][mid]:
            continue
        machine = cur['machines'][mid]
        machine['id'] = mid
        stream(('machine', 'change', machine))

    # relation add
    for rid in cur_rel_ids - prev_rel_ids:
        scope = cur['topology'].get_relation_scope(rid)
        interface = cur['topology'].get_relation_type(rid)
        services = cur['topology'].get_relation_services(rid)

        services = [
            (cur['topology'].get_service_name(_sid),
             services[_sid]) for _sid in services.keys()]
        relation = {
            'id': rid,
            'interface': interface,
            'scope': scope,
            'endpoints': services}
        stream(('relation', 'add', relation))

    for mid in cur_machine_ids - prev_machine_ids:
        machine = cur['machines'][mid]
        machine['id'] = mid
        stream(('machine', 'add', machine))

    # We want the added units to be sent in ascending order by number.
    for uid in sorted(set(cur_units) - set(prev_units), key=_unit_key):
        unit = cur_units[uid]
        unit['id'] = uid
        stream(('unit', 'add', unit))

    for uid in (set(cur_units).intersection(set(prev_units))):
        if cur_units[uid] == prev_units[uid]:
            continue
        unit = cur_units[uid]
        unit['id'] = uid
        stream(('unit', 'change', unit))

    for uid in (set(prev_units) - set(cur_units)):
        stream(('unit', 'remove', uid))

    for mid in prev_machine_ids - cur_machine_ids:
        stream(('machine', 'remove', mid))

    for rid in prev_rel_ids - cur_rel_ids:
        # with svc ids
        stream(('relation', 'remove', rid))

    for sid in prev_svc_ids - cur_svc_ids:
        stream(('service', 'remove', sid))

    for sid in cur_svc_ids.intersection(prev_svc_ids):
        svc = dict(cur['services'][sid])
        svc.pop('units', None)
        svc.pop('relations', None)

        prev_svc = dict(prev['services'][sid])
        prev_svc.pop('units', None)
        prev_svc.pop('relations', None)

        if svc != prev_svc:
            svc['id'] = sid
            stream(('service', 'change', svc))
