"""

additional context commands for state representation.
"""

from twisted.internet.defer import inlineCallbacks, returnValue

from juju.state.charm import CharmStateManager
from juju.state.service import ServiceStateManager
from juju.state.relation import RelationStateManager


@inlineCallbacks
def get_charm(context, charm_id):
    """
    GET /charms/:id

    Charms have embedded '/'

    Return back charm metadata.
    """

    charms = CharmStateManager(context.client)
    charm = yield charms.get_charm_state(charm_id)
    config = yield charm.get_config()
    md = yield charm.get_metadata()

    returnValue({
        'name': charm.name,
        'revision': charm.revision,
        'url': charm.id,
        'config': {'options': config.as_dict()},
        'provides': md.provides,
        'requires': md.requires,
        'peers': md.peers,
        'subordinate': md.is_subordinate,
        })


@inlineCallbacks
def get_service(context, service_name):
    """
    get('/service/foobar')
    set('/service/foobar/', value)

    GET /service/:id
    returns:

    - Service Info
    - Config
    - Constraints
    - Relations
    - Charm identifier.

    Set Service -> (context api set config, set constraints)
    """
    services = ServiceStateManager(context.client)
    service = yield services.get_service_state(service_name)
    config = yield service.get_config()
    constraints = yield service.get_constraints()
    charm_id = yield service.get_charm_id()
    relations = RelationStateManager(context.client)
    exposed = yield service.get_exposed_flag()

    rels = []
    for r in (yield relations.get_relations_for_service(service)):
        endpoint = None
        if r.relation_role != "peer":
            for s in (yield r.get_service_states()):
                if s.service_name != service.service_name:
                    endpoint = s.service_name
        rels.append(
            dict(name=r.relation_name,
                 ident=r.relation_ident,
                 role=r.relation_role,
                 scope=r.relation_scope,
                 endpoint=endpoint))

    returnValue({
        'name': service.service_name,
        'charm': charm_id,
        'config': dict(config),
        'exposed': exposed,
        'constraints': constraints.data,
        'rels': rels
        })


@inlineCallbacks
def get_endpoints(context, service_names=()):
    service_manager = ServiceStateManager(context.client)
    if not service_names:
        services = yield service_manager.get_all_service_states()
    else:
        services = []
        for s in service_names:
            services.append(
                (yield service_manager.get_service_state(s)))

    endpoint_mapping = {}

    for svc in services:
        charm = yield svc.get_charm_state()
        metadata = yield charm.get_metadata()
        endpoint_mapping[svc.service_name] = svc_rel_meta = {}
        svc_rel_meta['provides'] = _flatter(metadata.provides)
        svc_rel_meta['requires'] = _flatter(metadata.requires)

    returnValue(endpoint_mapping)


def _flatter(rel_meta):
    flattened = []
    if not rel_meta:
        return flattened

    for k in rel_meta:
        rel = dict(rel_meta[k])
        rel['name'] = k
        flattened.append(rel)
    return flattened
