from twisted.internet.defer import inlineCallbacks
from juju.rapi.tests.common import ContextTestBase


class APIContextTest(ContextTestBase):

    @inlineCallbacks
    def test_terminate_machine(self):
        yield self.mock_store_charms(["precise/mysql"])
        self.mocker.replay()

        yield self.context.deploy("precise/mysql", num_units=2)
        yield self.context.destroy_service("mysql")

        response = yield self.context.terminate_machine([1])
        self.assertEqual(response['result'], 1)
        self.assertEqual(
            response['log'],
            [('info', 'Machines terminated: 1')])

        response = yield self.context.status()
        self.assertEqual(
            response['result'],
            {'services': {},
             'machines': {0: {'instance-id': 'pending'}}})

    @inlineCallbacks
    def test_error(self):
        response = yield self.context.add_unit('garbage', 1)
        self.assertTrue(response['err'])
        self.assertTrue('failure' in response)

    @inlineCallbacks
    def test_deploy_and_destroy(self):
        yield self.mock_store_charms(["precise/mysql"])
        self.mocker.replay()

        response = yield self.context.deploy("precise/mysql")
        self.assertEqual(response['result'], True)
        self.assertEqual(
            [('info',
             'Deploying service mysql using charm cs:precise/mysql-1'),
             ('info',
              "Charm deployed as service: 'mysql'")],
            response['log'])

        response = yield self.context.destroy_service("mysql")
        self.assertEqual(response['result'], True)
        self.assertEqual(
            [('info', "Service 'mysql' destroyed.")],
            response['log'])

    @inlineCallbacks
    def test_add_and_remove_unit(self):
        yield self.mock_store_charms(["precise/wordpress"])
        self.mocker.replay()

        yield self.context.deploy("wordpress")
        response = yield self.context.add_unit('wordpress', 2)
        self.assertEqual(
            [('info',
              "Unit 'wordpress/1' added to service 'wordpress'"),
             ('info',
              "Unit 'wordpress/2' added to service 'wordpress'")],
            response['log'])
        self.assertEqual(response['result'], ['wordpress/1', 'wordpress/2'])

        response = yield self.context.remove_units(['wordpress/2'])
        self.assertEqual(
            [('info',
              "Unit 'wordpress/2' removed from service 'wordpress'")],
            response['log'])
        self.assertEqual(
            response['result'], True)

    @inlineCallbacks
    def test_add_and_remove_relation(self):
        yield self.mock_store_charms(["precise/wordpress", "precise/mysql"])
        self.mocker.replay()

        yield self.context.deploy("precise/mysql")
        yield self.context.deploy("wordpress")
        response = yield self.context.add_relation(
            'mysql', 'wordpress')
        self.assertIn(
            ('info', u'Added mysql relation to all service units.'),
            response['log'])
        self.assertEqual(
            response['result'],
            {'endpoints': [
                {'mysql': {'name': 'server', 'role': 'server'}},
                {'wordpress': {'name': 'db', 'role': 'client'}}],
             'id': 'relation-0000000000',
             'interface': 'mysql',
             'scope': 'global'})

        response = yield self.context.remove_relation(
            'mysql', 'wordpress')
        self.assertEqual(
            [('info', 'Removed mysql relation from all service units.')],
            response['log'])

    @inlineCallbacks
    def test_get_and_set_config(self):
        yield self.mock_store_charms(["precise/wordpress"])
        self.mocker.replay()

        yield self.context.deploy("wordpress")
        response = yield self.context.get_config('wordpress')

        config = {
            'charm': 'cs:precise/wordpress-3',
            'service': 'wordpress',
            'settings': {
                'blog-title': {
                    'default': True,
                    'type': 'string',
                    'value': 'My Title',
                    'description': 'A descriptive title used for the blog.'}}}
        self.assertEqual(config, response['result'])

        response = yield self.context.set_config(
            'wordpress', {'blog-title': 'my blog'})
        self.assertEqual(response['result'], True)

        del config['settings']['blog-title']['default']
        config['settings']['blog-title']['value'] = 'my blog'

        response = yield self.context.get_config('wordpress')
        self.assertEqual(config, response['result'])

    @inlineCallbacks
    def test_get_and_set_constraints(self):
        yield self.mock_store_charms(["precise/wordpress"])
        self.mocker.replay()

        yield self.context.deploy(
            "wordpress", constraints=['mem=4G', 'cpu=2'])

        cs = {'wordpress': {'mem': 4096.0,
                            'ubuntu-series': 'precise',
                            'arch': 'amd64',
                            'provider-type': 'dummy',
                            'cpu': 2.0}}
        response = yield self.context.get_constraints(['wordpress'])
        self.assertEqual(cs, response['result'])
        self.assertEqual(
            [('info', 'Fetching constraints for service wordpress')],
            response['log'])

        response = yield self.context.set_constraints(
            'wordpress', ['mem=8G', 'cpu=4'])
        self.assertEqual(response['result'], True)

        cs['wordpress']['mem'] = 1024 * 8
        cs['wordpress']['cpu'] = 4.0

        response = yield self.context.get_constraints(['wordpress'])
        self.assertEqual(response['result'], cs)

    @inlineCallbacks
    def test_expose_and_unexpose(self):
        yield self.mock_store_charms(["precise/wordpress"])
        self.mocker.replay()

        yield self.context.deploy("wordpress")
        response = yield self.context.status()

        stat = {'services': {
                'wordpress': {'units': {
                    'wordpress/0': {'machine': 0,
                                    'agent-state': 'pending',
                                    'public-address': None}},
                              'charm': 'cs:precise/wordpress-3',
                              'relations': {}}},
             'machines': {0: {'instance-id': 'pending'}}}
        self.assertEqual(response['result'], stat)

        response = yield self.context.expose('wordpress')
        self.assertTrue(response['result'])
        self.assertEqual(
            response['log'],
            [('info', "Service 'wordpress' was exposed.")])

        response = yield self.context.status()
        self.assertEqual(
            response['result']['services']['wordpress'].get(
                'exposed'), True)

        response = yield self.context.unexpose('wordpress')
        self.assertTrue(response['result'])
        self.assertEqual(
            response['log'],
            [('info', "Service 'wordpress' was unexposed.")])

        response = yield self.context.status()
        self.assertEqual(response['result'], stat)
