from twisted.internet.defer import inlineCallbacks, succeed

from juju.state.topology import InternalTopology
from juju.lib.testing import TestCase
from juju.rapi.delta import delta, DeltaStreamManager
from juju.rapi.tests.common import ContextTestBase


class DeltaTestBase(TestCase):

    counter = 0

    def get_changes(self, prev, cur):
        results = []
        delta(prev, cur, results.append)
        return results

    def add_relation(
        self, topology, iface,
        end_a, end_a_name, end_a_role,
        end_b=None, end_b_name=None, end_b_role=None,
        scope='global'):
        """ """
        self.counter += 1
        rid = "r-%s" % self.counter
        topology.add_relation(rid, iface, scope)

        sid_a = topology.find_service_with_name(end_a)
        if not sid_a:
            self.counter += 1
            sid_a = "s-%s" % self.counter
            topology.add_service(sid_a, end_a)

        topology.assign_service_to_relation(
            rid, sid_a, end_a_name, end_a_role)

        if end_b:
            assert end_b and end_b_name and end_b_role
        else:
            return

        sid_b = topology.find_service_with_name(end_b)
        if not sid_b:
            self.counter += 1
            sid_b = "s-%s" % self.counter
            topology.add_service(sid_b, end_b)

        topology.assign_service_to_relation(
            rid, sid_b, end_b_name, end_b_role)


class DeltaStreamManagerTest(ContextTestBase, DeltaTestBase):

    @inlineCallbacks
    def setUp(self):
        yield super(DeltaStreamManagerTest, self).setUp()
        self.stream = DeltaStreamManager(self.context)
        self.stream._schedule = lambda: 1
        yield self.client.create("/topology", InternalTopology().dump())
        self.changes = []
        self.stream.add(self)

    @inlineCallbacks
    def tearDown(self):
        self.stream.stop()
        self.context.close()
        yield super(DeltaStreamManagerTest, self).tearDown()

    # test act like a protocol/context to collect deltas.
    def publish_changes(self, changes):
        self.changes = changes

    def mock_current(self, status):
        self.mock_context = self.mocker.patch(self.stream.context)
        self.mock_context.status()
        self.mocker.result(succeed(status))

    @inlineCallbacks
    def test_delta_protocol(self):
        self.mock_current({
            'result': {
                'machines': {
                    0: {"instance-id": 0,
                        "instance-state": "running",
                        'agent-state': "running",
                        "dns-name": "steam-cloud-1.com"}},
                'services': {
                    'mysql':
                    {"charm": "local:series/mysql-1",
                     "relations": {"db": ["wordpress"]},
                     "units": {
                         "mysql/0": {
                             "machine": 0,
                             "public-address": "mysql-0.example.com",
                             "agent-state": "running"}}}},
                }})
        self.mocker.replay()
        yield self.stream.start()
        self.assertEqual(self.changes, [
            ('service', 'add', {'charm': 'local:series/mysql-1',
                                'id': 'mysql'}),
             ('machine', 'add', {'agent-state': 'running',
                                 'dns-name': 'steam-cloud-1.com',
                                 'instance-id': 0,
                                 'id': 0,
                                 'instance-state': 'running'}),
             ('unit', 'add', {'agent-state': 'running',
                              'machine': 0,
                              'id': 'mysql/0',
                              'public-address': 'mysql-0.example.com'})])

    def test_get_current_state_empty(self):
        self.assertEqual(
            self.stream.get_current(),
            [])
        yield self.stream.start()
        self.assertEqual(
            self.stream.get_current(),
            [])

    def xtest_start_with_state(self):
        self.mock_start()
        self.mocker.replay()
        self.assertEqual(
            self.stream.get_current(),
            [])
        return self.stream.start()


class DeltaTest(DeltaTestBase):

    # additional delta tests in test_ws

    def test_delta_1(self):
        """delete a service and watches the ripples.
        """
        prev_topology = InternalTopology()
        self.add_relation(
            prev_topology,
            'rdb', 'mysql', 'db', 'client', 'wordpress', 'db', 'server')

        cur_topology = InternalTopology()

        prev = {
            'machines': {
                0: {"instance-id": 0,
                    "instance-state": "running",
                    'agent-state': "running",
                    "dns-name": "steam-cloud-1.com"},
                1: {"instance-id": 1,
                    "instance-state": "running",
                    'agent-state': "pending",
                    "dns-name": "steam-cloud-2.com"},
                },
            'services': {
                'mysql':
                {"charm": "local:series/mysql-1",
                 "relations": {"db": ["wordpress"]},
                 "units": {
                     "mysql/0": {
                         "machine": 0,
                         "public-address": "mysql-0.example.com",
                         "agent-state": "running"}}},
                'wordpress':
                {"charm": "local:series/wordpress-1",
                 "relations": {"db": ["wordpress"]},
                 "units": {
                     "memcache/0": {
                         "machine": 1,
                         "public-address": "wordpress-0.example.com",
                         "agent-state": "pending"}}}},
            'topology': prev_topology
            }

        cur = {
            'machines': {
                0: {"instance-id": 0,
                    "instance-state": "running",
                    'agent-state': "running",
                    "dns-name": "steam-cloud-1.com"},
                },
            'services': {
                'mysql':
                {"charm": "local:series/mysql-1",
                 "units": {
                     "mysql/0": {
                         "machine": 0,
                         "public-address": "mysql-0.example.com",
                         "agent-state": "running"}}},
                },
            'topology': cur_topology
            }

        changes = self.get_changes(prev, cur)
        self.assertEqual(
            [('unit', 'remove', 'memcache/0'),
             ('machine', 'remove', 1),
             ('relation', 'remove', 'r-1'),
             ('service', 'remove', 'wordpress')],
            changes)

    def test_delta_2(self):
        """
        """
        prev_topology = InternalTopology()
        self.add_relation(prev_topology, 'riak', 'riak', 'ring', 'peer')

        cur_topology = InternalTopology()
        cur_topology.parse(prev_topology.dump())
        self.add_relation(
            cur_topology,
            'riak', 'riak', 'client', 'server', 'myapp', 'db', 'client')

        prev = {
            'machines': {
                0: {"instance-id": 0,
                    "instance-state": "running",
                    'agent-state': "running",
                    "dns-name": "steam-cloud-1.com"},
                },
            'services': {
                'riak':
                {"charm": "local:series/riak-1",
                 "relations": {"ring": ["riak"]},
                 "units": {
                     "riak/0": {
                         "machine": 0,
                         "public-address": "mysql-0.example.com",
                         "agent-state": "running"}}}},
            'topology': prev_topology}

        cur = {
            'machines': {
                0: {"instance-id": 0,
                    "instance-state": "running",
                    'agent-state': "running",
                    "dns-name": "steam-cloud-1.com"},
                1: {"instance-id": 1,
                    "instance-state": "running",
                    'agent-state': "running",
                    "dns-name": "steam-cloud-2.com"},
                2: {"instance-id": 2,
                    "instance-state": "running",
                    'agent-state': "pending",
                    "dns-name": "steam-cloud-3.com"},
                3: {"instance-id": 3,
                    "instance-state": "running",
                    'agent-state': "pending",
                    "dns-name": "steam-cloud-4.com"},
                },
            'services': {
                'riak':
                {"charm": "local:series/riak-1",
                 "relations": {"ring": ["riak"], "client": ["myapp"]},
                 "units": {
                     "riak/1": {
                         "machine": 1,
                         "public-address": "riak-1.example.com",
                         "agent-state": "running"},
                     "riak/2": {
                         "machine": 2,
                         "public-address": "riak-2.example.com",
                         "agent-state": "pending"}}},
                'myapp':
                {"charm": "local:series/myapp-1",
                 "relations": {"db": ["riak"]},
                 "units": {
                     "myapp/0": {
                         "machine": 3,
                         "public-address": "myapp-0.example.com",
                         "agent-state": "running"}}}},
            'topology': cur_topology
            }

        changes = self.get_changes(prev, cur)
        self.assertEqual(
            [('service', 'add', {'id': 'myapp',
                                 'charm': 'local:series/myapp-1'}),
             ('relation', 'add',
              {'endpoints': [('myapp', {'name': 'db', 'role': 'client'}),
                             ('riak', {'name': 'client', 'role': 'server'})],
               'id': 'r-3',
               'interface': 'riak',
               'scope': 'global'}),
             ('machine', 'add',
              {'agent-state': 'running',
               'dns-name': 'steam-cloud-2.com',
               'instance-id': 1,
               'id': 1,
               'instance-state': 'running'}),
             ('machine', 'add',
              {'agent-state': 'pending',
               'dns-name': 'steam-cloud-3.com',
               'instance-id': 2,
               'id': 2,
               'instance-state': 'running'}),
             ('machine', 'add',
              {'agent-state': 'pending',
               'dns-name': 'steam-cloud-4.com',
               'instance-id': 3,
               'id': 3,
               'instance-state': 'running'}),
             ('unit', 'add',
              {'agent-state': 'running',
               'machine': 3,
               'id': 'myapp/0',
               'public-address': 'myapp-0.example.com'}),
             ('unit', 'add',
              {'agent-state': 'running',
               'machine': 1,
               'id': 'riak/1',
               'public-address': 'riak-1.example.com'}),
             ('unit', 'add',
              {'agent-state': 'pending',
               'id': 'riak/2',
               'machine': 2,
               'public-address': 'riak-2.example.com'}),
             ('unit', 'remove', 'riak/0')],
            changes)

    def xtest_delta_3(self):
        # test subordinates deltas
        pass

    def xtest_delta_4(self):
        cur = InternalTopology()
        changes = self.get_changes(
            {'services': {}, 'machines': {}, 'topology': None},
            {'machines': {
                0: {"instance-id": 0,
                    "instance-state": "running",
                    'agent-state': "running",
                    "dns-name": "steam-cloud-1.com"},
                },
             'services': {
                 'riak':
                 {"charm": "local:series/riak-1",
                  "units": {
                      "riak/0": {
                          "machine": 0,
                          "public-address": "mysql-0.example.com",
                          "agent-state": "running"}}}},
            'topology': cur})
        self.assertEqual(
            [('service', 'add', {'charm': 'local:series/riak-1'}),
             ('machine', 'add', {'agent-state': 'running',
                                 'instance-state': 'running',
                                 'instance-id': 0,
                                 'dns-name': 'steam-cloud-1.com'}),
             ('unit', 'add', {'machine': 0, 'agent-state': 'running',
                              'public-address': 'mysql-0.example.com'})],
            changes)

    def test_delta_sorts_added_units_by_name_and_number(self):
        cur = InternalTopology()
        changes = self.get_changes(
            {'services': {}, 'machines': {}, 'topology': None},
            {'machines': {
                0: {"instance-id": 0,
                    "instance-state": "running",
                    'agent-state': "running",
                    "dns-name": "steam-cloud-1.com"},
                1: {"instance-id": 1,
                    "instance-state": "running",
                    'agent-state': "running",
                    "dns-name": "steam-cloud-2.com"},
                2: {"instance-id": 2,
                    "instance-state": "running",
                    'agent-state': "running",
                    "dns-name": "steam-cloud-3.com"},
                3: {"instance-id": 3,
                    "instance-state": "running",
                    'agent-state': "running",
                    "dns-name": "steam-cloud-4.com"},
                4: {"instance-id": 4,
                    "instance-state": "running",
                    'agent-state': "running",
                    "dns-name": "steam-cloud-5.com"},
                },
             'services': {
                 'riak':
                 {"charm": "local:series/riak-1",
                  "units": {
                      "riak/1": {
                          "machine": 0,
                          "public-address": "riak-3.example.com",
                          "agent-state": "running"},
                      "riak/0": {
                          "machine": 1,
                          "public-address": "riak-2.example.com",
                          "agent-state": "running"},
                      "riak/3": {
                          "machine": 2,
                          "public-address": "riak-0.example.com",
                          "agent-state": "pending"},
                      "riak/2": {
                          "machine": 3,
                          "public-address": "riak-1.example.com",
                          "agent-state": "stopping"},
                      "mysql/4": {
                          "machine": 4,
                          "public-address": "mysql-1.example.com",
                          "agent-state": "running"}}}},
            'topology': cur})
        unit_ids = [data['id'] for category, action, data in changes
                    if (category, action)==('unit', 'add')]
        self.assertEqual(
            ['mysql/4', 'riak/0', 'riak/1', 'riak/2', 'riak/3'],
            unit_ids)