import json
import logging


from twisted.internet.defer import inlineCallbacks, Deferred

from juju.rapi.transport.ws import WebSocketAPIFactory
from juju.providers.dummy import MachineProvider

from juju.rapi.tests.common import ContextTestBase
from juju.state.topology import InternalTopology
from juju.state.tests.common import StateTestBase
from juju.tests.common import get_test_zookeeper_address


class OWTransport(object):
    """A writeable, observable transport for tests.
    """
    def __init__(self, protocol):
        self.stream = []
        self.watches = []
        self.closed = Deferred()
        self.protocol = protocol

    def watch(self, filter):
        d = Deferred()
        self.watches.append((d, filter))
        return d

    def watch_kv(self, key, value):
        def _f(data):
            return data.get(key) == value
        return self.watch(_f)

    def watch_dict(self, matching):
        def _f(data):
            found = True
            for k in matching:
                if data.get(k) != matching[k]:
                    found = False
                    break
            return found
        return self.watch(_f)

    def write(self, data):
        msg = json.loads(data)
        self.stream.append(msg)
        for d, f in self.watches:
            if f(msg):
                d.callback(msg)

    def writeSequence(self, data_set):
        for d in data_set:
            self.write(d)

    def loseConnection(self):
        self.protocol.connectionLost()
        self.closed.callback(True)

    def getPeer(self):
        raise NotImplemented()

    def getHost(self):
        raise NotImplemented()


class WebSocketTestBase(StateTestBase):

    @inlineCallbacks
    def setUp(self):
        yield super(WebSocketTestBase, self).setUp()
        self.output = self.capture_logging(
            "juju.rapi.ws", level=logging.DEBUG)
        self.provider = MachineProvider(
            "cloudy", {"default-series": "precise"})
        self.ws_factory = WebSocketAPIFactory(
            get_test_zookeeper_address(), self.provider)
        self.ws = self.ws_factory.buildProtocol()
        self.ws.transport = self.transport = OWTransport(self.ws)
        self.ws._schedule = lambda: 1

    @inlineCallbacks
    def tearDown(self):
        if self.ws.active:
            self.ws.connectionLost()
        yield self.ws.finished
        yield super(WebSocketTestBase, self).tearDown()


class WebSocketTest(WebSocketTestBase):

    def test_greeting(self):

        d = self.transport.watch_dict(dict(
            version=0, ready=True, provider_type="dummy",
            default_series="precise"))
        self.ws.connectionMade()
        return d

    @inlineCallbacks
    def test_malformed_message(self):
        d = self.transport.watch_dict(dict(
            err=True, ready=False))
        finished = self.ws.connectionMade()
        self.assertEqual(self.ws_factory.numClients, 1)
        self.ws.dataReceived("abc 123")
        yield d
        yield self.transport.closed
        self.assertEqual(self.ws_factory.numClients, 0)
        yield finished

    @inlineCallbacks
    def test_status(self):
        d = self.transport.watch_kv('op', 'status')
        self.ws.connectionMade()
        self.ws.dataReceived(json.dumps(dict(
            op='status')))
        msg = yield d
        self.assertEqual(msg['result']['services'], {})
        self.assertEqual(msg['result']['machines'], {})


class DeltaSocketTest(WebSocketTestBase, ContextTestBase):

    @inlineCallbacks
    def setUp(self):
        yield super(DeltaSocketTest, self).setUp()
        yield self.set_topology(InternalTopology())

        # No background timer, we'll explicitly
        # pump and collect messages.
        self.stream_manager = self.ws_factory.stream_manager
        self.stream_manager._schedule = lambda: 1
        yield self.ws_factory.startFactory()

    #@inlineCallbacks
    def tearDown(self):
        self.ws_factory.stopFactory()
        return super(DeltaSocketTest, self).tearDown()

    @inlineCallbacks
    def test_delta_stream(self):
        yield self.mock_store_charms(["precise/wordpress", "precise/mysql"])
        self.mocker.replay()

        # There's always an implicit initial pump of current state made
        # on connection. But no initial state for this test.
        self.ws.connectionMade()

        # Modify state and send events out.
        yield self.context.deploy("precise/mysql")
        yield self.context.deploy("wordpress")
        yield self.stream_manager.pump()

        delta = self.transport.stream[-1]
        self.assertEqual(delta['op'], 'delta')
        self.assertEqual(delta['result'], [
            [u'service', u'add', {u'id': 'wordpress',
                                  u'charm': u'cs:precise/wordpress-3'}],
            [u'service', u'add', {u'id': 'mysql',
                                  u'charm': u'cs:precise/mysql-1'}],
            [u'machine', u'add', {'id': 0, u'instance-id': u'pending'}],
            [u'machine', u'add', {'id': 1, u'instance-id': u'pending'}],
            [u'unit', u'add', {u'machine': 0, u'agent-state': u'pending',
                               u'public-address': None, 'id': u'mysql/0'}],
            [u'unit', u'add', {u'machine': 1, u'agent-state': u'pending',
                               u'public-address': None, 'id': u'wordpress/0'}],
            ])

        # Try it again, but first get the current state dumped as changes.
        changes = self.stream_manager.get_current()
        self.assertEqual(len(changes), 6)
        # Tuple vs. list doesn't matter in json, normalize for assertion.
        for x, y in zip(delta['result'], changes):
            self.assertEqual(tuple(x), y)

        yield self.context.add_relation("wordpress", "mysql")
        yield self.stream_manager.pump()
        delta = self.transport.stream[-1]
        self.assertEqual(
            delta['result'], [[
                u'relation', u'add', {
                    u'interface': u'mysql',
                    u'scope': u'global',
                    u'endpoints': [
                        [u'mysql', {u'role': u'server', u'name': u'server'}],
                        [u'wordpress', {u'role': u'client', u'name': u'db'}]],
                    u'id': u'relation-0000000000'}]],
            )

        # Once more with some deletions.
        yield self.context.destroy_service('mysql')
        yield self.stream_manager.pump()
        delta = self.transport.stream[-1]
        self.assertEqual(
            delta['result'],
            [[u'unit', u'remove', u'mysql/0'],
             [u'relation', u'remove', u'relation-0000000000'],
             [u'service', u'remove', u'mysql']
            ])

    @inlineCallbacks
    def xtest_delta_stream_initial_state(self):
        yield self.ws_factory.startFactory()
        self.ws.connectionMade()
