import json
import logging

from twisted.internet.defer import inlineCallbacks, Deferred, DeferredQueue
from twisted.internet.protocol import Factory, Protocol

from txzookeeper.managed import ManagedClient

from juju.rapi.context import APIContext
from juju.rapi.delta import DeltaStreamManager


class MalformedMessage(Exception):
    """An invalid message recieved."""


log = logging.getLogger("juju.rapi.ws")


class APIWebSocket(Protocol):
    """
    Juju Websocket API

    The client MUST wait for initial handshake from the server.
    """
    def __init__(self, zk_hosts, provider):
        # Are we currently accepting/processing messages.
        self.active = False
        # Zk servers to connect to.
        self.zk_hosts = zk_hosts
        # Env Provider (needed for status and constraint set retrieval).
        self.provider = provider
        # API Context/Endpoint
        self.context = None
        # Client to zk.
        self.client = None
        # Operation queue
        self.queue = None
        # Deferred marker that the web socket resources are finalized/closed.
        self.finished = None

        self.last_status = None

    @inlineCallbacks
    def process_pending(self, *ignored):
        """Process the next message off the websocket channel.
        """
        self.transport.write(json.dumps(
            {"version": 0, "extensions": [], "ready": True,
             "provider_type": self.provider.provider_type,
             "default_series": self.provider.config.get("default-series")}))

        log.info("Sent greeting")

        # Push the current state.
        self.publish_changes()

        while 1:
            if not self.active:
                break
            msg = yield self.queue.get()
            if msg is None:
                break
            yield self.process_msg(msg)

        self.finished.callback(True)

#    def publish_msg(self, msg):
#        if not self.context or not self.active:
#            return
#        self.transport.write(json.dumps(msg))

    def publish_changes(self, changes=None):
        if not self.context or not self.active:
            return
        msg = {'op': 'delta'}
        if changes is None:
            changes = self.factory.stream_manager.get_current()

        if not changes:
            return

        msg['result'] = changes
        log.debug("send delta update %s" % changes)

        # TODO: Chunk this data?
        self.transport.write(json.dumps(msg))

    @inlineCallbacks
    def process_msg(self, msg):
        """Process an operation msg, and send the result.

        msg structure (json serialized dict)
          expected keys
           - 'request_id' / a passthrough value.
           - 'op' / specify the command to run.
        """
        try:
            data = json.loads(msg)
        except ValueError:
            raise MalformedMessage(msg)

        if not isinstance(data, dict):
            raise MalformedMessage(msg)

        log.info("Process message %s" % msg)

        # Ghetto object publishing, todo whitelist.
        op = data.get('op', None)
        if not op or not op[0].isalpha():
            raise MalformedMessage(msg)
        rid = data.pop('request_id', None)
        method = getattr(self.context, data.pop('op'))
        log.info("Invoking method %s with %s" % (method, data))
        result = yield method(**data)

        data.update(result)
        data['op'] = op
        if rid is not None:
            data['request_id'] = rid
        log.info("Returning results %s" % data)
        self.transport.write(json.dumps(data))

    def handle_msg_error(self, err):
        # TODO: return back a request/op structure with error info.
        log.warning("Error %s" % err)
        self.active = False
        self.transport.write(json.dumps(
            {"err": True, "ready": False}))
        self.transport.loseConnection()
        self.finished.callback(False)

    def handle_close(self, *ignored):
        self.client.close()
        self.context = None
        self.queue = None
        log.debug("Closed")

    # Protocol methods

    def connectionMade(self):
        self.client = ManagedClient(self.zk_hosts)
        self.context = APIContext(self.client, self.provider)
        self.queue = DeferredQueue()
        self.finished = Deferred()
        self.active = True

        self.factory.numClients += 1
        log.debug("Connection made. Clients:%s" % self.factory.numClients)
        d = self.client.connect()
        d.addCallback(self.process_pending)
        d.addErrback(self.handle_msg_error)
        return d

    def dataReceived(self, data):
        log.debug("Data received %s" % data)
        self.queue.put(data)

    def connectionLost(self, reason=None):
        self.active = False
        self.queue.put(None)
        self.factory.numClients -= 1
        self.factory.stream_manager.remove(self)
        log.debug("Connection Lost. Clients:%s" % self.factory.numClients)
        self.finished.addCallback(self.handle_close)


class WebSocketAPIFactory(Factory):

    protocol = APIWebSocket

    def __init__(self, zk_hosts, provider):
        self.zk_hosts = zk_hosts
        self.provider = provider
        self.numClients = 0
        self.stream_manager = DeltaStreamManager(
            APIContext(ManagedClient(self.zk_hosts), self.provider))

    # The common usage for twisted's native ws support is via protocol
    # wrapper mounted onto an http server. This usage prevents
    # startFactory/stopFactory from being invoked automatically. Users
    # of this factory should invoke these two methods appropriatly.

    def startFactory(self):
        log.debug("Initialized ws factory")
        return self.stream_manager.start()

    def stopFactory(self):
        log.debug("Stopped ws factory")
        return self.stream_manager.stop()

    # New connection.
    def buildProtocol(self, addr=None):
        p = self.protocol(self.zk_hosts, self.provider)
        p.factory = self
        self.stream_manager.add(p)
        return p
